/*
 * Copyright 2008 Evan Cowden
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.nakedosgi.processor.scr;

import java.util.List;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.MirroredTypeException;
import javax.tools.Diagnostic;

import org.nakedosgi.Constants;
import org.nakedosgi.annotations.NoValue;
import org.nakedosgi.annotations.scr.SCRBind;
import org.nakedosgi.processor.model.Component;
import org.nakedosgi.processor.model.Reference;

/**
 * Parser for the {@link SCRBind} annotation
 * 
 * @author Evan Cowden
 */
public class SCRBindParser {
    private ProcessingEnvironment processingEnv;

    /**
     * Create a new instance
     * @param processingEnv
     */
    public SCRBindParser(ProcessingEnvironment processingEnv) {
        this.processingEnv = processingEnv;
    }

    /**
     * Search a class for {@link SCRBind} annotations and add to a component
     * correspondingly
     * 
     * @param clazz
     * the class to search
     * @param component
     * the component to modify
     */
    public void parse(TypeElement clazz, Component component) {
        List<? extends Element> elements = clazz.getEnclosedElements();
        for (Element e : elements) {
            SCRBind annotation = e.getAnnotation(SCRBind.class);
            if (annotation != null) {
                ExecutableElement method = (ExecutableElement) e;
                handle(annotation, method, component);
            }

        }
    }

    /**
     * Parse a single annotation on a single method, altering the component
     * accordingly
     * @param annotation
     * @param method
     * @param component
     */
    void handle(SCRBind annotation, ExecutableElement method, Component component) {

        // Verify that the method has exactly one parameter
        if (BindUnbindUtil.parametersAreInvalid(method)) {
            processingEnv.getMessager().printMessage(Diagnostic.Kind.ERROR,
                    "SCR binding methods must have exactly one parameter", method);
            return;
        }

        // Get all necessary objects
        // ...annotation & name
        String annotationName = annotation.name();
        String name = BindUnbindUtil.getName(method, annotationName);

        // ...Model components
        Reference reference = BindUnbindUtil.getReference(name, component);

        // Copy info over from annotation
        // ...cardinality
        reference.setCardinality(annotation.cardinality());

        // ...interface
        // (This is where I start thinking that the person that designed this
        // API needs to be shot...although it is a pretty smart hack...)
        String declaredClassName = null;
        try {
            annotation.interface_().getName(); // Will throw
                                                // MirroredTypeException
        } catch (MirroredTypeException ex) {
            declaredClassName = ex.getTypeMirror().toString();
        }
        if (declaredClassName.equals(NoValue.class.getName())) {
            VariableElement parameter = method.getParameters().get(0);
            reference.setInterface(parameter.asType().toString());
        } else {
            reference.setInterface(declaredClassName);
        }

        // ...policy
        reference.setPolicy(annotation.policy());

        // ...target
        if (!annotation.target().equals(Constants.NO_VALUE)) {
            reference.setTarget(annotation.target());
        }

        // ...bind
        String bind = method.getSimpleName().toString();
        reference.setBind(bind);
    }
}
