/*
 * Copyright 2008 Evan Cowden
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.nakedosgi.processor.scr;

import static org.nakedosgi.Constants.NO_VALUE;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.ProcessingEnvironment;
import javax.annotation.processing.Processor;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.MirroredTypeException;
import javax.lang.model.type.TypeKind;
import javax.lang.model.type.TypeMirror;
import javax.tools.Diagnostic.Kind;
import javax.xml.bind.JAXBException;

import org.nakedosgi.annotations.NoValue;
import org.nakedosgi.annotations.NullableBoolean;
import org.nakedosgi.annotations.scr.SCRComponent;
import org.nakedosgi.annotations.scr.SCRReference;
import org.nakedosgi.processor.ParsingUtil;
import org.nakedosgi.processor.model.Component;
import org.nakedosgi.processor.model.Implementation;
import org.nakedosgi.processor.model.Provide;
import org.nakedosgi.processor.model.Reference;
import org.nakedosgi.processor.model.Service;
import org.nakedosgi.spi.ServiceProvider;

/**
 * Processor for the {@link SCRComponent} annotation
 * @author Evan Cowden
 */
@ServiceProvider(Processor.class)
@SupportedSourceVersion(SourceVersion.RELEASE_6)
@SupportedAnnotationTypes("org.nakedosgi.annotations.scr.*")
public class SCRComponentProcessor extends AbstractProcessor {
    private SCRBindParser bindParser;
    private SCRUnbindParser unbindParser;

    // Map between component classes and components
    private Map<String, Component> components = new HashMap<String, Component>();

    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
	super.init(processingEnv);
	bindParser = new SCRBindParser(processingEnv);
	unbindParser = new SCRUnbindParser(processingEnv);
    }

    @Override
    public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv) {
	Set<? extends Element> elements = roundEnv.getElementsAnnotatedWith(SCRComponent.class);
	processingEnv.getMessager().printMessage(Kind.NOTE, "Performing Naked-SCR Processing...");

	for (Element element : elements) {
	    Component component = SCRProcessingUtil.getComponent(element, components);

	    // Get info
	    TypeElement type = (TypeElement) element; // Safe since
	    // SCRComponent has a
	    // @Target restraint
	    SCRComponent annotation = type.getAnnotation(SCRComponent.class);

	    // Parse top-level component info
	    // ...Implementation class
	    Implementation implementation = new Implementation();
	    String qualifiedName = ParsingUtil.getQualifiedName(type.asType());
	    implementation.setClazz(qualifiedName);
	    component.setImplementation(implementation);

	    // ...enabled
	    component.setEnabled(annotation.enabled());

	    // ...name
	    if (annotation.name().equals(NO_VALUE)) {
		component.setName(type.getSimpleName().toString());
	    } else {
		component.setName(annotation.name());
	    }

	    // ...factory
	    if (!annotation.factory().equals(NO_VALUE)) {
		component.setFactory(annotation.factory());
	    }

	    // ...immediate
	    if (!annotation.immediate().equals(NullableBoolean.UNSPECIFIED)) {
		component.setImmediate(annotation.immediate().value());
	    }

	    // ...serviceFactory
	    if (annotation.serviceFactory() == true) {
		// ...create new Service if one does not exist
		Service service = component.getService();
		if (service == null) {
		    service = new Service();
		    component.setService(service);
		}

		service.setServicefactory(annotation.serviceFactory());
	    }

	    // ...interfaces
	    List<String> interfaces = new ArrayList<String>();
	    /*
	     * Find all interfaces, put them here, then add to component model
	     * later
	     */
	    TypeMirror iface;
	    try {
		@SuppressWarnings("unused")
		Object willThrowExceptionHere = annotation.iface();
	    } catch (MirroredTypeException e) {
		iface = e.getTypeMirror();
		String ifaceName = ParsingUtil.getQualifiedName(iface);
		if (ifaceName.equals(Object.class.getName())) {
		    // Object is default, search for interfaces
		    List<? extends TypeMirror> declaredInterfaces = type.getInterfaces();
		    for (TypeMirror typeMirror : declaredInterfaces) {
			interfaces.add(typeMirror.toString());
		    }
		} else if (ifaceName.equals(NoValue.class.getName())) {
		    // NoValue specified, Do Nothing
		} else {
		    // Interface specified, add it
		    interfaces.add(ifaceName);
		}
	    }

	    // ......then add to the component model
	    if (interfaces.size() > 0) {
		Service service = component.getService();
		if (service == null) {
		    service = new Service();
		    component.setService(service);
		}
		for (String ifaceName : interfaces) {
		    Provide provide = new Provide();
		    provide.setInterface(ifaceName);
		    service.getProvide().add(provide);
		}
	    }

	    // ...references
	    for (SCRReference scrReference : annotation.references()) {
		Reference reference = new Reference(scrReference);
		component.getReference().add(reference);
	    }

	    runChildParsers(component, type);

	    String className = type.getQualifiedName().toString();
	    components.put(className, component);
	}

	// Write configuration when complete
	if (roundEnv.processingOver()) {
	    processingDone();
	}

	return true;
    }

    /**
     * Run child parsers, such as binding parsers, recursively searching
     * superclasses
     * @param component
     * @param type
     */
    private void runChildParsers(Component component, TypeElement type) {
	TypeMirror superType = type.getSuperclass();
	if (superType.getKind().equals(TypeKind.DECLARED)) {
	    DeclaredType superDeclared = (DeclaredType) superType;
	    TypeElement superElement = (TypeElement) superDeclared.asElement();
	    runChildParsers(component, superElement);
	}

	bindParser.parse(type, component);
	unbindParser.parse(type, component);
    }

    public void processingDone() {
	// Write configuration files
	try {
	    SCRProcessingUtil.writeComponentDescriptors(components.values(), processingEnv);
	} catch (JAXBException ex) {
	    // TODO
	    processingEnv.getMessager().printMessage(Kind.ERROR,
		    "An error occured while writing Naked-SCR configuration");
	    ex.printStackTrace();
	} catch (IOException ex) {
	    // TODO
	    processingEnv.getMessager().printMessage(Kind.ERROR,
		    "An error occured while writing Naked-SCR configuration");
	    ex.printStackTrace();
	}
    }

    /**
     * Get the specified component out of the components collection. If it does
     * not already exist, create it and add it to the collection.
     * @param componentClass
     * @param components
     * @return
     */
    protected static Component getComponent(String componentClass, Map<String, Component> components) {
	Component result = components.get(componentClass);

	if (result == null) {
	    result = new Component();
	    components.put(componentClass, result);
	}

	return result;
    }
}
