/*
 * Copyright 2008 Evan Cowden
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.nakedosgi.processor.scr;

import java.util.List;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;
import javax.tools.Diagnostic;

import org.nakedosgi.annotations.scr.SCRUnbind;
import org.nakedosgi.processor.model.Component;
import org.nakedosgi.processor.model.Reference;

/**
 * Processor for the {@link SCRUnbind} annotation
 * @author Evan Cowden
 */
public class SCRUnbindParser {
    private ProcessingEnvironment processingEnv;

    /**
     * Create a new instance
     * @param processingEnv
     */
    public SCRUnbindParser(ProcessingEnvironment processingEnv) {
        this.processingEnv = processingEnv;
    }

    /**
     * Search a class for {@link SCRUnbind} annotations and add to a component
     * correspondingly
     * 
     * @param clazz
     * the class to search
     * @param component
     * the component to modify
     */
    public void parse(TypeElement clazz, Component component) {
        List<? extends Element> elements = clazz.getEnclosedElements();
        for (Element e : elements) {
            SCRUnbind annotation = e.getAnnotation(SCRUnbind.class);
            if (annotation != null) {
                ExecutableElement method = (ExecutableElement) e;
                handle(annotation, method, component);
            }

        }
    }

    /**
     * Parse a single annotation on a single method, altering the component accordingly
     * @param annotation
     * @param method
     * @param component
     */
    void handle(SCRUnbind annotation, ExecutableElement method, Component component) {

        // Verify that the method has exactly one parameter
        if (BindUnbindUtil.parametersAreInvalid(method)) {
            processingEnv.getMessager().printMessage(Diagnostic.Kind.ERROR,
                    "SCR binding methods must have exactly one parameter", method);
            return;
        }

        // Get all necessary objects
        // ...annotation & name
        String annotationName = annotation.name();
        String name = BindUnbindUtil.getName(method, annotationName);

        // ...Model components
        Reference reference = BindUnbindUtil.getReference(name, component);

        // ...bind
        String unbind = method.getSimpleName().toString();
        reference.setUnbind(unbind);
    }

}
